//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Library General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include "action.h"
#include <stdlib.h>
#include <strstream>
#include "stack.h"

#define debug(x) {std::cerr<<__FILE__<<": "<<__LINE__<<": "<<x<<endl;}

Action::Action()
{
}

Action::~Action()
{
}

Action* Action::handle_package(data_package* mydata)
{
	Action* a;
	switch (mydata->action)
	{
		case ACTION_STACK_FIGHT:
			a = new Action_Fight(mydata);
			break;

		case ACTION_STACK_SPLIT:
			a = new Action_Split(mydata);
			break;

		case ACTION_STACK_FIGHT_CITY:
			a = new Action_Fight_City(mydata);
			break;

		case ACTION_STACK_JOIN:
			a = new Action_Join(mydata);
			break;

		case ACTION_STACK_MOVED_ONE_STEP:
			a = new Action_Move_Step(mydata);
			break;

		case ACTION_CITY_OCCUPY:
			a = new Action_Occupy(mydata);
			break;

		case ACTION_CITY_RAZE:
			a = new Action_Raze(mydata);
			break;
			
		case ACTION_STACK_SEARCH_RUIN:
			a = new Action_Search(mydata);
			break;

		case ACTION_STACK_VISIT_TEMPLE:
			a =new Action_Temple(mydata);
			break;

		case ACTION_REWARD:
			a = new Action_Reward(mydata);
			break;

		case ACTION_CITY_UPGRADE_DEFENSE:
			a = new Action_Upgrade_Defense(mydata);
			break;

		case ACTION_CITY_BUY_PRODUCTION:
			a = new Action_Buy_Production(mydata);
			break;

		case ACTION_CITY_CHANGE_PRODUCTION:
			a = new Action_Change_Production(mydata);
			break;
	}
	return a;
}

//-----------------------------------------------------------------------------
//Action_Fight

Action_Fight::Action_Fight()
	:Action()
{
	d_id = ACTION_STACK_FIGHT;

	for (int i =0; i < 8; i++)
		d_attacker_damages[i] = d_defender_damages[i] = 0;
}

Action_Fight::Action_Fight(data_package* mydata)
{
	std::list<int>::iterator it = mydata->data.begin();

	d_attacker_id = (unsigned int) (*it);	it++;
	d_defender_id = (unsigned int) (*it);	it++;

	for (int i = 0; i < 8; i++, it++)
		d_attacker_damages[i] = (*it);
	
	for (int i = 0; i < 8; i++, it++)
		d_defender_damages[i] = (*it);
}


Action_Fight::~Action_Fight()
{
}

std::string Action_Fight::dump()
{
	std::strstream s;
	s << "Attacker ID: " << d_attacker_id;
	s << "\nDefender ID: " << d_defender_id;

	s << "\nAttacker causalties:\n";
	for (int i = 0; i < 8; i++)
		s << d_attacker_damages[i] << " ";

	s << "\nDefender causalties:\n";
	for (int i = 0; i < 8; i++)
		s << d_defender_damages[i] << " ";
	s << "\n";

	return s.str();
}

data_package Action_Fight::get_data()
{
	data_package mydata;
	mydata.action = ACTION_STACK_FIGHT;

	mydata.data.push_back((int)(d_attacker_id));
	mydata.data.push_back((int)(d_defender_id));

	for (int i = 0; i < 8; i++)
		mydata.data.push_back(d_attacker_damages[i]);
	
	for (int i = 0; i < 8; i++)
		mydata.data.push_back(d_defender_damages[i]);

	return mydata;
}

bool Action_Fight::fill_data(Stack* attacker, Stack* defender,
				Stack* orig_attacker, Stack* orig_defender)
{
	if ((attacker->size() != orig_attacker->size())
		|| (defender->size() != orig_defender->size())
		|| (attacker->size() > 8) || (attacker->empty())
		|| (defender->size() > 8) || (defender->empty()))
	{
		std::cerr <<"Action_Fight::fill_data(): wrong stack size\n";
		return false;
	}

	debug("Action_Fight::fill_data()")
	
	d_attacker_id = orig_attacker->get_id();
	d_defender_id = orig_defender->get_id();

	Stack::iterator it, orig_it;

	it = attacker->begin();
	orig_it = orig_attacker->begin();

	for (int i = 0; it != attacker->end(); i++, it++, orig_it++)
		d_attacker_damages[i] = (*orig_it)->get_hp() - (*it)->get_hp();	
	
	it = defender->begin();
	orig_it = orig_defender->begin();

	for (int i = 0; it != defender->end(); i++, it++, orig_it++)
		d_defender_damages[i] = (*orig_it)->get_hp() - (*it)->get_hp();
	
	return true;
}

//-----------------------------------------------------------------------------
//Action_Split

Action_Split::Action_Split()
	:Action()
{
	d_id = ACTION_STACK_SPLIT;
	
	for (int i = 0; i < 8; i++)
		d_armies_moved[i] = 0;
}

Action_Split::Action_Split(data_package* mydata)
{
	d_id = ACTION_STACK_SPLIT;

	std::list<int>::iterator it = mydata->data.begin();
	
	d_orig_stack_id = static_cast<unsigned int>(*it);		it++;
	d_added_stack_id = static_cast<unsigned int>(*it);		it++;

	for (int i = 0; i < 8; i++, it++)
		d_armies_moved[i] = static_cast<unsigned int>(*it);
}	

Action_Split::~Action_Split()
{
}

std::string Action_Split::dump()
{
	std::strstream s;
	s <<"Stack " <<d_orig_stack_id; 
	s <<" forked new stack " <<d_added_stack_id <<".\n";
	s <<"moved armies:\n";

	for (int i = 0; i < 8; i++)
		s <<d_armies_moved[i] <<"  ";
	
	s <<"\n";
	return s.str();
}

data_package Action_Split::get_data()
{
	data_package mydata;

	mydata.action = ACTION_STACK_SPLIT;
	mydata.data.push_back(static_cast<int>(d_orig_stack_id));
	mydata.data.push_back(static_cast<int>(d_added_stack_id));

	for (int i = 0; i < 8; i++)
		mydata.data.push_back(static_cast<int>(d_armies_moved[i]));
	
	return mydata;
}

bool Action_Split::fill_data(Stack* orig_stack, Stack* added_stack)
{
	if ((orig_stack->size() > 8) ||(added_stack->size() > 8)
		|| (orig_stack->empty()) || (added_stack->empty()))
	{
		std::cerr <<"Action_Split::fill_data(): wrong stack size\n";
		return false;
	}
	
	debug("Action_Split::fill_data()")

	d_orig_stack_id = orig_stack->get_id();
	d_added_stack_id = added_stack->get_id();

	Stack::iterator it = added_stack->begin();
	for (int i = 0; it != added_stack->end(); it++, i++)
		d_armies_moved[i] = (*it)->get_id();
	
	return true;
}

//-----------------------------------------------------------------------------
//Action_Join

Action_Join::Action_Join()
	:Action()
{
	d_id = ACTION_STACK_JOIN;
	d_orig_id = d_joining_id = 0;
}

Action_Join::Action_Join(data_package* mydata)
{
	d_id = ACTION_STACK_JOIN;
	d_orig_id = *(mydata->data.begin());
	d_joining_id = *((mydata->data.begin())++);
}

Action_Join::~Action_Join()
{
}

std::string Action_Join::dump()
{
	strstream s;
	s <<"Stack " <<d_joining_id <<" joined Stack " <<d_orig_id <<".\n";
	return s.str();
}

data_package Action_Join::get_data()
{
	data_package mydata;

	mydata.action = ACTION_STACK_JOIN;
	mydata.data.push_back(static_cast<unsigned int>(d_orig_id));
	mydata.data.push_back(static_cast<unsigned int>(d_joining_id));

	return mydata;
}

bool Action_Join::fill_data(Stack* orig, Stack* joining)
{
	if ((orig->empty()) || (joining->empty())
		|| (orig->size() + joining->size() > 8))
	{
		std::cerr <<"Action_Join::fill_data(): wrong stack size\n";
		return false;
	}
	
	debug("Action_Join::fill_data")
	
	d_orig_id = orig->get_id();
	d_joining_id = joining->get_id();
	return true;
}

//-----------------------------------------------------------------------------
//Action_Fight_City

Action_Fight_City::Action_Fight_City()
	:Action()
{
	d_id = ACTION_STACK_FIGHT_CITY;

	for (int i = 0; i < Fight::ATTACKER_MAX; i++)
	{
		d_attacker_ids[i] = 0;
		for (int k = 0; k < 8; k++)
			d_attacker_damages[i][k] = 0;
	}

	for (int i = 0; i < Fight::DEFENDER_MAX; i++)
	{
		d_defender_ids[i] = 0;
		for (int k = 0; k < 8; k++)
			d_defender_damages[i][k] = 0;
	}
}

Action_Fight_City::Action_Fight_City(data_package* mydata)
	:Action()
{
	d_id = ACTION_STACK_FIGHT_CITY;
	std::list<int>::iterator it = mydata->data.begin();

	for (int i = 0; i < Fight::ATTACKER_MAX; i++, it++)
		d_attacker_ids[i] = static_cast<unsigned int>(*it);
	
	for (int i = 0; i < Fight::DEFENDER_MAX; i++, it++)
		d_defender_ids[i] = static_cast<unsigned int>(*it);
	
	for (int i = 0; i < Fight::ATTACKER_MAX; i++)
		for (int k = 0; k < 8; k++, it++)
			d_attacker_damages[i][k] = (*it);
	
	for (int i = 0; i < Fight::DEFENDER_MAX; i++)
		for (int k = 0; k < 8;k++, it++)
			d_defender_damages[i][k] = (*it);
}

Action_Fight_City::~Action_Fight_City()
{
}

std::string Action_Fight_City::dump()
{
	std::strstream s;
	int a = 0;
	int d = 0;
	
	s <<"The stacks ";
	while ((a < Fight::ATTACKER_MAX) && (d_attacker_ids[a] != 0))
	{
		s <<d_attacker_ids[a] <<" ";
		a++;
	}
	
	s <<"fought against city occupied by stacks ";
	while ((d < Fight::DEFENDER_MAX) && (d_defender_ids[d] != 0))
	{
		s <<d_defender_ids[d] <<" ";
		d++;
	}
	
	s<<"\nDamages on attacker side:\n";
	for (int i = 0; i < a; i++)
	{
		for (int k = 0; k < 8; k++)
			s <<d_attacker_damages[i][k] <<" ";
		s <<endl;
	}

	s<<"\nDamages on defender side:\n";
	for (int i = 0; i < d; i++)
	{
		for (int k = 0; k < 8; k++)
			s <<d_defender_damages[i][k] <<" ";
		s <<endl;
	}

	return s.str();
}

data_package Action_Fight_City::get_data()
{
	data_package mydata;

	mydata.action = ACTION_STACK_FIGHT_CITY;
	
	for (int i = 0; i < Fight::ATTACKER_MAX; i++)
		mydata.data.push_back(static_cast<int>(d_attacker_ids[i]));
	
	for (int i = 0; i < Fight::DEFENDER_MAX; i++)
		mydata.data.push_back(static_cast<int>(d_defender_ids[i]));
	
	for (int i = 0; i < Fight::ATTACKER_MAX; i++)
		for (int k = 0; k < 8; k++)
			mydata.data.push_back(d_attacker_damages[i][k]);
	
	for (int i = 0; i < Fight::DEFENDER_MAX; i++)
		for (int k = 0; k < 8; k++)
			mydata.data.push_back(d_defender_damages[i][k]);
	
	return mydata;
}

bool Action_Fight_City::fill_data(std::vector<Stack*> orig_attackers,
					std::vector<Stack*> orig_defenders,
					std::vector<Stack*> attackers,
					std::vector<Stack*> defenders)
{
	if ((orig_attackers.size() > (unsigned int)Fight::ATTACKER_MAX)
		|| (orig_attackers.size() != attackers.size())
		|| (orig_defenders.size() > (unsigned int)Fight::DEFENDER_MAX)
		|| (orig_defenders.size() != defenders.size()))
	{
		std::cerr <<"Action_Fight_City::fill_data: wrong vector size\n";
		return false;
	}

	debug("Action_Fight_City::fill_data()")

	std::vector<Stack*>::iterator it, it2;
	Stack::iterator sit, sit2;

	//calculate attacker damages etc.
	it = orig_attackers.begin();
	it2 = attackers.begin();
	for (int i = 0;	it != orig_attackers.end(); it++, it2++, i++)
	{
		d_attacker_ids[i] = (*it)->get_id();
		sit = (*it)->begin();
		sit2 = (*it2)->begin();
		for (int k = 0; sit != (*it)->end(); sit++, sit2++, k++)
			d_attacker_damages[i][k] = (*sit)->get_hp() - (*sit2)->get_hp();
	}

	//now the same for the defenders
	it = orig_defenders.begin();
	it2 = defenders.begin();
	for (int i = 0;	it != orig_defenders.end(); it++, it2++, i++)
	{
		d_defender_ids[i] = (*it)->get_id();
		sit = (*it)->begin();
		sit2 = (*it2)->begin();
		for (int k = 0; sit != (*it)->end(); sit++, sit2++, k++)
			d_defender_damages[i][k] = (*sit)->get_hp() - (*sit2)->get_hp();
	}

	return true;
}

//-----------------------------------------------------------------------------
//Action_Move_Step

Action_Move_Step::Action_Move_Step()
	:Action(), d_stack_id(0)
{
	d_id = ACTION_STACK_MOVED_ONE_STEP;
	d_dest.x = d_dest.y = 0;
}

Action_Move_Step::Action_Move_Step(data_package* mydata)
	:Action()
{
	d_id = ACTION_STACK_MOVED_ONE_STEP;
	std::list<int>::iterator it = mydata->data.begin();
	
	d_stack_id = static_cast<unsigned int>(*it);	it++;
	d_dest.x = (*it);				it++;
	d_dest.y = (*it);				it++;
}

Action_Move_Step::~Action_Move_Step()
{
}

std::string Action_Move_Step::dump()
{
	strstream s;
	s <<"Stack " <<d_stack_id;
	s <<" moved to " <<d_dest.x <<"," <<d_dest.y <<endl;
	return s.str();
}

data_package Action_Move_Step::get_data()
{
	data_package mydata;
	
	mydata.action = ACTION_STACK_MOVED_ONE_STEP;
	mydata.data.push_back(static_cast<int>(d_stack_id));
	mydata.data.push_back(d_dest.x);
	mydata.data.push_back(d_dest.y);

	return mydata;
}

bool Action_Move_Step::fill_data(Stack* s, PG_Point dest)
{
	d_stack_id = s->get_id();
	d_dest = dest;
	return true;
}

//-----------------------------------------------------------------------------
//Action_Occupy

Action_Occupy::Action_Occupy()
	:Action(), d_city_id(0)
{
	d_id = ACTION_CITY_OCCUPY;
}

Action_Occupy::Action_Occupy(data_package* mydata)
{
	d_id = ACTION_CITY_OCCUPY;
	d_city_id = static_cast<unsigned int>(*(mydata->data.begin()));
}

Action_Occupy::~Action_Occupy()
{
}

std::string Action_Occupy::dump()
{
	std::strstream s;
	s <<"City " <<d_city_id <<" occupied\n";
	return s.str();
}

data_package Action_Occupy::get_data()
{
	data_package mydata;
	mydata.action = ACTION_CITY_OCCUPY;
	mydata.data.push_back(static_cast<int>(d_city_id));
	return mydata;
}

bool Action_Occupy::fill_data(City* c)
{
	d_city_id = c->get_id();
	return true;
}

//-----------------------------------------------------------------------------
//Action_Raze

Action_Raze::Action_Raze()
	:Action(), d_city_id(0)
{
	d_id = ACTION_CITY_RAZE;
}

Action_Raze::Action_Raze(data_package* mydata)
	:Action()
{
	d_id = ACTION_CITY_RAZE;
	d_city_id = static_cast<unsigned int>(*(mydata->data.begin()));
}

Action_Raze::~Action_Raze()
{
}

std::string Action_Raze::dump()
{
	std::strstream s;
	s <<"City " <<d_city_id <<" razed\n";
	return s.str();
}

data_package Action_Raze::get_data()
{
	data_package mydata;
	mydata.action = ACTION_CITY_RAZE;
	mydata.data.push_back(static_cast<int>(d_city_id));
	return mydata;
}

bool Action_Raze::fill_data(City* c)
{
	d_city_id = c->get_id();
	return true;
}

//-----------------------------------------------------------------------------
//Action_Search

Action_Search::Action_Search()
	:Action(), d_ruin_id(0), d_stack_id(0), d_searched(0)
{
	d_id = ACTION_STACK_SEARCH_RUIN;
	for (int i = 0; i < 8; i++)
	{
		d_keeper_damages[i] = 0;
		d_seeker_damages[i] = 0;
	}
}

Action_Search::Action_Search(data_package* mydata)
{
	d_id = ACTION_STACK_SEARCH_RUIN;
	std::list<int>::iterator it = mydata->data.begin();
	
	d_ruin_id = static_cast<unsigned int>(*it);	it++;
	d_stack_id = static_cast<unsigned int>(*it);	it++;
	d_searched = static_cast<bool>(*it);		it++;

	for (int i = 0; i < 8; i++)
	{
		d_keeper_damages[i] = (*it);
		it++;
	}

	for (int i = 0; i < 8; i++)
	{
		d_seeker_damages[i] = (*it);
		it++;
	}
}

Action_Search::~Action_Search()
{
}

std::string Action_Search::dump()
{
	std::strstream s;
	s << "Ruin number " <<d_ruin_id <<"searched by stack " <<d_stack_id;
	s <<"\nLosses of keepers:\n";
	for (int i = 0; i < 8; i++)
		s <<d_keeper_damages[i] <<" ";
	s <<"\n\nLosses of explorers:\n";
	for (int i = 0; i < 8; i++)
		s <<d_seeker_damages[i] <<" ";
	s <<"\nRuin has been searched: " <<d_searched <<endl;
	return s.str();
}

data_package Action_Search::get_data()
{
	data_package mydata;
	mydata.action = ACTION_STACK_SEARCH_RUIN;
	mydata.data.push_back(static_cast<int>(d_ruin_id));
	mydata.data.push_back(static_cast<int>(d_stack_id));
	mydata.data.push_back(static_cast<int>(d_searched));
	for (int i = 0; i < 8; i++)
		mydata.data.push_back(d_keeper_damages[i]);
	for (int i = 0; i < 8; i++)
		mydata.data.push_back(d_seeker_damages[i]);
	
	return mydata;
}

bool Action_Search::fill_data(Ruin* r,Stack* orig_keepers, Stack*orig_explorers,
				Stack* explorers, bool searched)
{
	d_ruin_id = r->get_id();
	d_stack_id = explorers->get_id();
	d_searched = searched;

	//no fight, so we are finished here
	if ((orig_keepers == 0) || (explorers == 0) || (orig_explorers == 0))
		return true;

	Stack::iterator it, it2;
	it = r->get_occupant()->begin();
	it2 = orig_keepers->begin();

	for (int i = 0; it2 != orig_keepers->end(); it++, it2++, i++)
		d_keeper_damages[i] = (*it2)->get_hp() - (*it)->get_hp();

	it = explorers->begin();
	it2 = orig_explorers->begin();
	for (int i = 0; it2 != orig_explorers->end(); it++, it2++, i++)
		d_seeker_damages[i] = (*it2)->get_hp() - (*it)->get_hp();
	
	return true;
}

//-----------------------------------------------------------------------------
//Action_Temple

Action_Temple::Action_Temple()
	:Action(), d_temple_id(0), d_stack_id(0)
{
	d_id = ACTION_STACK_VISIT_TEMPLE;
}

Action_Temple::Action_Temple(data_package* mydata)
	:Action()
{
	d_id = ACTION_STACK_VISIT_TEMPLE;

	std::list<int>::iterator it = mydata->data.begin();

	d_temple_id = static_cast<unsigned int>(*it);	it++;
	d_stack_id = static_cast<unsigned int>(*it);
}

Action_Temple::~Action_Temple()
{
}

std::string Action_Temple::dump()
{
	std::strstream s;
	s<<"Stack #" <<d_stack_id <<" visited temple #" <<d_temple_id <<".\n";
	return s.str();
}

data_package Action_Temple::get_data()
{
	data_package mydata;
	
	mydata.action = ACTION_STACK_VISIT_TEMPLE;
	mydata.data.push_back(static_cast<int>(d_temple_id));
	mydata.data.push_back(static_cast<int>(d_stack_id));

	return mydata;
}

bool Action_Temple::fill_data(Temple* t, Stack* s)
{
	d_temple_id = t->get_id();
	d_stack_id = s->get_id();

	return true;
}

//-----------------------------------------------------------------------------
//Action_Upgrade_Defense

Action_Upgrade_Defense::Action_Upgrade_Defense()
	:Action(), d_city_id(0)
{
	d_id = ACTION_CITY_UPGRADE_DEFENSE;
}

Action_Upgrade_Defense::Action_Upgrade_Defense(data_package* mydata)
	:Action()
{
	d_id = ACTION_CITY_UPGRADE_DEFENSE;
	d_city_id = static_cast<unsigned int>(*(mydata->data.begin()));
}

Action_Upgrade_Defense::~Action_Upgrade_Defense()
{
}

std::string Action_Upgrade_Defense::dump()
{
	std::strstream s;
	s <<"Defenses upgraded of city #" <<d_city_id <<".\n";

	return 	s.str();
}

data_package Action_Upgrade_Defense::get_data()
{
	data_package mydata;
	
	mydata.action = ACTION_CITY_UPGRADE_DEFENSE;
	mydata.data.push_back(static_cast<int>(d_city_id));

	return mydata;
}

bool Action_Upgrade_Defense::fill_data(City* c)
{
	d_city_id = c->get_id();
	return true;
}

//-----------------------------------------------------------------------------
//Action_Buy_Production

Action_Buy_Production::Action_Buy_Production()
	:Action(), d_city_id(0), d_prod(0)
{
	d_id = ACTION_CITY_BUY_PRODUCTION;
}

Action_Buy_Production::Action_Buy_Production(data_package* mydata)
	:Action()
{
	d_id = ACTION_CITY_BUY_PRODUCTION;
	std::list<int>::iterator it = mydata->data.begin();

	d_city_id = static_cast<unsigned int>(*it);	it++;
	d_prod = (*it);
}

Action_Buy_Production::~Action_Buy_Production()
{
}

std::string Action_Buy_Production::dump()
{
	std::strstream s;
	s <<"production #" <<d_prod <<" bought in city #" <<d_city_id <<".\n";

	return s.str();
}

data_package Action_Buy_Production::get_data()
{
	data_package mydata;
	
	mydata.action = ACTION_CITY_BUY_PRODUCTION;
	mydata.data.push_back(static_cast<int>(d_city_id));
	mydata.data.push_back(d_prod);

	return mydata;
}

bool Action_Buy_Production::fill_data(City* c, int prod)
{
	d_city_id = c->get_id();
	d_prod = prod;

	return true;
}

//-----------------------------------------------------------------------------
//Action_Change_Production

Action_Change_Production::Action_Change_Production()
	:Action(), d_city_id(0), d_prod(0)
{
	d_id = ACTION_CITY_CHANGE_PRODUCTION;
}

Action_Change_Production::Action_Change_Production(data_package* mydata)
{
	d_id = ACTION_CITY_CHANGE_PRODUCTION;

	std::list<int>::iterator it = mydata->data.begin();

	d_city_id = static_cast<unsigned int>(*it);	it++;
	d_prod = (*it);
}

Action_Change_Production::~Action_Change_Production()
{
}

std::string Action_Change_Production::dump()
{
	std::strstream s;
	s <<"Production of city #" <<d_city_id <<" changed to #" <<d_prod <<".\n";

	return s.str();
}

data_package Action_Change_Production::get_data()
{
	data_package mydata;

	mydata.action = ACTION_CITY_CHANGE_PRODUCTION;
	mydata.data.push_back(static_cast<int>(d_city_id));
	mydata.data.push_back(d_prod);

	return mydata;
}

bool Action_Change_Production::fill_data(City* c, int prod)
{
	d_city_id = c->get_id();
	d_prod = prod;

	return true;
}

//-----------------------------------------------------------------------------
//Action_Reward

Action_Reward::Action_Reward()
	:Action(), d_stack_id(0)
{
	d_id = ACTION_REWARD;
	d_reward.type = 0;
}

Action_Reward::Action_Reward(data_package* mydata)
{
	d_id = ACTION_REWARD;
	std::list<int>::iterator it = mydata->data.begin();

	d_stack_id = static_cast<unsigned int>(*it);		it++;
	d_reward.type = static_cast<unsigned int>(*it);		it++;
	d_reward.subtype = static_cast<unsigned int>(*it);	it++;
	d_reward.amount = (*it);
}

Action_Reward::~Action_Reward()
{
}

std::string Action_Reward::dump()
{
	std::strstream s;
	s <<"reward given to stack number " <<d_stack_id <<endl;
	s <<"Type: " <<d_reward.type <<", Subtype: " <<d_reward.subtype;
	s <<", Amount: " <<d_reward.amount <<endl;

	return s.str();
}

data_package Action_Reward::get_data()
{
	data_package mydata;
	mydata.action = ACTION_REWARD;

	mydata.data.push_back(static_cast<int>(d_stack_id));
	mydata.data.push_back(static_cast<int>(d_reward.type));
	mydata.data.push_back(static_cast<int>(d_reward.subtype));
	mydata.data.push_back(d_reward.amount);

	return mydata;
}

bool Action_Reward::fill_data(Stack* s, fl_Treasure reward)
{
	if (s)
		d_stack_id = s->get_id();
	d_reward = reward;

	return true;
}
