//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Library General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifndef ACTION_H
#define ACTION_H

#include <string>
#include "defs.h"
#include "stack.h"
#include "fight.h"
#include "city.h"
#include "ruin.h"
#include "temple.h"

/* For each action the players can do, there is an appropriate data structure
 * to hold what the player has done in which city etc. We define an abstract
 * class here and derive different classes for different actions (as they need
 * different input). */

class Action
{
	public:
		Action();
		virtual ~Action();

		virtual std::string dump()=0;	//return debug info
		virtual data_package get_data()=0; //return the data so
						//it can be sent via network
		static Action* handle_package (data_package* mydata);

		unsigned int get_id() {return d_id;}
		
	//bool fill_data();		each derived class has a function of
	//				this type, but with different arguments,
	//				so there is not really any use of having
	//				this function in the base class
	
	protected:
		unsigned int d_id;
};

//-----------------------------------------------------------------------------

class Action_Fight : public Action
{
	public:
		Action_Fight();
		Action_Fight(data_package* mydata);
		~Action_Fight();

		std::string dump();
		data_package get_data();

		bool fill_data(Stack* attacker, Stack* defender,
				Stack* orig_attacker, Stack* orig_defender);

	private:
		unsigned int d_attacker_id, d_defender_id;
		int d_attacker_damages[8], d_defender_damages[8];
};

//-----------------------------------------------------------------------------

class Action_Split : public Action
{
	public:
		Action_Split();
		Action_Split(data_package* mydata);
		~Action_Split();

		std::string dump();
		data_package get_data();

		bool fill_data(Stack* orig_stack, Stack* added_stack);
			//both stacks are already split!!!
	
	private:
		unsigned int d_orig_stack_id, d_added_stack_id;
		unsigned int d_armies_moved[8];
};

//-----------------------------------------------------------------------------

class Action_Join : public Action
{
	public:
		Action_Join();
		Action_Join(data_package* mydata);
		~Action_Join();

		std::string dump();
		data_package get_data();
		bool fill_data(Stack* orig, Stack* joining);
	
	private:
		unsigned int d_orig_id, d_joining_id;
};

//-----------------------------------------------------------------------------

class Action_Fight_City : public Action
{
	public:
		Action_Fight_City();
		Action_Fight_City(data_package* mydata);
		~Action_Fight_City();

		std::string dump();
		data_package get_data();

		bool fill_data(std::vector<Stack*> orig_attackers,
				std::vector<Stack*> orig_defenders,
				std::vector<Stack*> attackers,
				std::vector<Stack*> defenders);
		
	private:
		unsigned int d_attacker_ids [Fight::ATTACKER_MAX];
		unsigned int d_defender_ids [Fight::DEFENDER_MAX];
		int d_attacker_damages [Fight::ATTACKER_MAX][8];
		int d_defender_damages [Fight::DEFENDER_MAX][8];
};

//-----------------------------------------------------------------------------

class Action_Move_Step : public Action
{
	public:
		Action_Move_Step();
		Action_Move_Step(data_package* mydata);
		~Action_Move_Step();

		std::string dump();
		data_package get_data();

		bool fill_data(Stack* s, PG_Point dest);
	
	private:
		unsigned int d_stack_id;
		PG_Point d_dest;
};

//-----------------------------------------------------------------------------

class Action_Occupy : public Action
{
	public:
		Action_Occupy();
		Action_Occupy(data_package* mydata);
		~Action_Occupy();

		std::string dump();
		data_package get_data();

		bool fill_data (City* c);
	
	private:
		unsigned int d_city_id;
};

//-----------------------------------------------------------------------------

class Action_Raze : public Action
{
	public:
		Action_Raze();
		Action_Raze(data_package* mydata);
		~Action_Raze();

		std::string dump();
		data_package get_data();

		bool fill_data (City* c);
	
	private:
		unsigned int d_city_id;
};
		
#endif //ACTION_H

//-----------------------------------------------------------------------------

class Action_Search : public Action
{
	public:
		Action_Search();
		Action_Search(data_package* mydata);
		~Action_Search();

		std::string dump();
		data_package get_data();

		bool fill_data(Ruin* r, Stack* orig_keepers,
				Stack* orig_explorers, Stack* explorers,
				bool searched);
	
	private:
		unsigned int d_ruin_id;
		unsigned int d_stack_id;
		bool d_searched;
		int d_keeper_damages[8];
		int d_seeker_damages[8];
};

//-----------------------------------------------------------------------------

class Action_Temple : public Action
{
	public:
		Action_Temple();
		Action_Temple(data_package* mydata);
		~Action_Temple();

		std::string dump();
		data_package get_data();

		bool fill_data(Temple* t, Stack* s);
	
	private:
		unsigned int d_temple_id;
		unsigned int d_stack_id;
};

//-----------------------------------------------------------------------------

class Action_Upgrade_Defense : public Action
{
	public:
		Action_Upgrade_Defense();
		Action_Upgrade_Defense(data_package* mydata);
		~Action_Upgrade_Defense();

		std::string dump();
		data_package get_data();

		bool fill_data(City* c);

	private:
		unsigned int d_city_id;
};

//-----------------------------------------------------------------------------

class Action_Buy_Production : public Action
{
	public:
		Action_Buy_Production();
		Action_Buy_Production(data_package* mydata);
		~Action_Buy_Production();

		std::string dump();
		data_package get_data();

		bool fill_data(City* c, int prod);

	private:
		unsigned int d_city_id;
		int d_prod;
};

//-----------------------------------------------------------------------------

class Action_Change_Production : public Action
{
	public:
		Action_Change_Production();
		Action_Change_Production(data_package* mydata);
		~Action_Change_Production();

		std::string dump();
		data_package get_data();

		bool fill_data(City* c, int prod);

	private:
		unsigned int d_city_id;
		int d_prod;
};

//-----------------------------------------------------------------------------

//this class will certainly have to be changed a lot, it doesn't handle anything
//else but gold, but I want at least a basic

class Action_Reward : public Action
{
	public:
		Action_Reward();
		Action_Reward(data_package* mydata);
		~Action_Reward();
		
		std::string dump();
		data_package get_data();

		bool fill_data (Stack* s, fl_Treasure reward);
	
	private:
		unsigned int d_stack_id;
		fl_Treasure d_reward;
};
