//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Library General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include "fight.h"
#include <assert.h>
#include <algorithm>
#include <stdlib.h>     // for random numbers
#include <math.h>       // for has_hit()
#include "army.h"
#include "stacklist.h"
#include "mapdef.h"
#include "citylist.h"
#include "city.h"
#include "stack.h"
#include "defs.h"

//#define debug(x) {std::cerr<<__FILE__<<": "<<__LINE__<<": "<<x<<endl;}
#define debug(x)

Fight::Fight(Stack* attacker, Stack* defender, bool def_bonus)
{
	assert(attacker->size() != 0);
	assert(defender->size() != 0);
	
	for (Stack::iterator it=attacker->begin(); it != attacker->end(); it++)
		d_attacker.push_back(*it);
	
	for (Stack::iterator it=defender->begin(); it != defender->end(); it++)
		d_defender.push_back(*it);
	
	d_add_defense = 1;
	d_open_fight = def_bonus;
	if (def_bonus)
		calculate_defense_bonus(defender);
	d_turn = 0;
	check_hero(d_attacker);
	check_hero(d_defender);
}

Fight::Fight(std::vector<Stack*>& attacker, std::vector<Stack*>& defender,
		bool def_bonus)
{
	Stack::iterator it;

	for (unsigned int i = 0; i < attacker.size(); i++)
		for (it=attacker[i]->begin(); it!=attacker[i]->end(); it++)
			d_attacker.push_back(*it);
	assert(!d_attacker.empty());
	
	for (unsigned int i = 0; i < defender.size(); i++)
		for(it=defender[i]->begin(); it!=defender[i]->end(); it++)
			d_defender.push_back(*it);
	assert(!d_defender.empty());

	d_add_defense = 1;
	if (def_bonus)
		calculate_defense_bonus(defender[0]);
	d_turn = 0;
	check_hero(d_attacker);
	check_hero(d_defender);
}

Fight::~Fight()
{
}

// calculate the general defense bonus for the defenders
void Fight::calculate_defense_bonus(Stack* stack)
{
	d_pos = stack->get_pos();	// used in fight_armies later

	if (mapdef->map[d_pos.x][d_pos.y]->get_building() == Maptile::CITY)
	{
		City* city = citylist->get_object_at(stack->get_pos());
		d_add_defense = 1 + (city->get_defense_level() * 0.2);
	}
	else
	{
		d_add_defense = ((*mapset)[mapdef->map[d_pos.x][d_pos.y]->get_type()]->get_defense()) / 10 + 1;
	}	
}


bool Fight::new_round()
{
	d_round_finished = false;
	d_attacker_finished = false;
	d_active = d_attacker.begin();
	d_turn++;

	if (d_turn > MAX_ROUNDS)
		return false;
	return true;
}


Fight::RESULT Fight::next_fight()
{
	debug("next_fight()");
	// if there haven't been any defender

	if (!d_attacker_finished)
	{
		int random = rand() % d_defender.size();

		if ((fight_armies(*d_active, d_defender[random], true))
			&& (d_defender[random]->get_hp() <= 0))
		{
			//defender army has been killed
			//this is for efficiency, deques are made to erase the
			//first/last item very fast
			d_defender[random] = d_defender[0];
			d_defender.erase(d_defender.begin());

			if (d_defender.empty())
				return ATTACKER_WON;

			check_hero (d_defender);
		}

		if (++d_active == d_attacker.end())
		{
			d_active = d_defender.begin();
			d_attacker_finished = true;
		}

		return NONE;
	}

	//else defender's turn
	int random =rand() % d_attacker.size();
	if ((fight_armies(*d_active, d_attacker[random], false))
		&& (d_attacker[random]->get_hp() <= 0))
	{
		//attacker army has been killed
		d_attacker[random] = d_attacker[0];
		d_attacker.erase(d_attacker.begin());

		if (d_attacker.empty())
			return DEFENDER_WON;

		check_hero(d_attacker);
	}

	if (++d_active == d_defender.end())
	{
		d_round_finished = true;
	}

	return NONE;
}


Fight::RESULT Fight::next_round()
{
	//this function is almost the same as next_fight, but a complete round
	//is done here

	int random;

	//attacker's turn
	for (std::deque<Army*>::iterator it = d_attacker.begin();
			it != d_attacker.end(); it++)
	{
		random = rand() % d_defender.size();
		if ((fight_armies(*it, d_defender[random], true))
			&& (d_defender[random]->get_hp() <= 0))
		{
			d_defender[random] = d_defender[0];
			d_defender.erase(d_defender.begin());

			if (d_defender.empty())
				return ATTACKER_WON;

			check_hero(d_defender);
		}
	}

	//defender's turn
	for (std::deque<Army*>::iterator it = d_defender.begin();
			it != d_attacker.end(); it++)
	{
		random = rand() % d_attacker.size();
		if ((fight_armies(*it, d_attacker[random], false))
			&& (d_attacker[random]->get_hp() <= 0))
		{
			d_attacker[random] = d_attacker[0];
			d_attacker.erase(d_attacker.begin());

			if (d_attacker.empty())
				return DEFENDER_WON;

			check_hero(d_attacker);
		}
	}

	return NONE;
}


bool Fight::fight_armies (Army* culprit, Army* victim, bool attack)
{
	debug("fight_armies()");

	unsigned int culprit_bonus = culprit->get_army_bonus();
	unsigned int victim_bonus = victim->get_army_bonus();


	float strength = culprit->get_strength();
	float defense = victim->get_defense();

	//add general defense bonus to victim
	if (attack)
	{
		defense *= d_add_defense;
	}

	//now comes the evaluation of several special army abilities as defined
	//in defs.h

	//if army is cavalry and open terrain and enemy not anticavalry, 
	//add +1 to strength
	//TBD: maptile type is only compared with MAPTILE_NONE, which could be  
	//a problem with other maptilesets than the standard one. Fix it!!
	//(don't know how to do it exactly)
	if (d_open_fight && (culprit_bonus & ARMY_CAVALRY) && attack
			&& (mapdef->map[d_pos.x][d_pos.y]->get_maptile_type() == MAPTILE_NONE)
			&& !(victim_bonus & ARMY_ANTICAVALRY))
		strength++;

	//if hero is on attackers side, increase strength
	if ((culprit_bonus & ARMY_HAS_HERO) && (!(culprit_bonus & ARMY_HERO)))
		strength++;
	
	//a hero on the side of the defender increases it's defense
	if ((victim_bonus & ARMY_HAS_HERO) && (!(victim_bonus & ARMY_HERO)))
		defense++;

	//if anticavalry unit and enemy is cavalry, double strength
	if ((culprit_bonus & ARMY_ANTICAVALRY) && (victim_bonus & ARMY_CAVALRY))
		strength *= 2;

	//unit can do instant kills with a chance of ... 1%
	if (culprit_bonus & ARMY_CRITICAL)
		if ((rand() % 100) == 0)
		{
			victim->damage(victim->get_hp());
			return true;
		}

	if (has_hit (strength, defense))
	{
		victim->damage(1);
		return true;
	}
	return false;
}

// The algorithm used here is as follows. We take a function
// f(x) = 50 * 0.75^(x)
// with x being the difference between attack strength and defense value.
// If the attack strength is smaller, the propability of the attacker hitting
// the defender is f(x), else it is (1 - f(x)). This has the advantage that
// the propability of a weak unit to hit a heavily armoured one steadily 
// goes towards zero with increasing inequality, but non-linear.

// There is always a limit of at least 5% /at most 95% to hit a unit (to
// give really weak units at least the chance to hit a stronger one before 
// being killed)

bool Fight::has_hit(float att_strength, float def_strength)
{
	int diff = abs ((int) (att_strength - def_strength));
	int chance = (int) (50 *(pow (0.75, diff)));
	if (chance < 5) chance = 5;
	else if (chance > 95) chance = 95;

	if (att_strength > def_strength) chance = 100 - chance;

	if ((rand() % 100) <= chance) return true;
	else return false;
}

//check and mark if there is a hero in the stack

void Fight::check_hero(std::deque<Army*>& fighters)
{
	bool has_hero = false;

	for (std::deque<Army*>::iterator it = fighters.begin();
		it != fighters.end(); it++)
	{
		if ((*it)->get_army_bonus() & ARMY_HERO)
		{
			has_hero = true;
			break;
		}
	}

	if (has_hero)
	{
		//set has_hero flag
		for (std::deque<Army*>::iterator it = fighters.begin();
				it != fighters.end(); it++)
			(*it)->set_army_bonus((*it)->get_army_bonus() | ARMY_HAS_HERO);
		return;
	}

	//mask out hero flag
	unsigned int compl_mask = !((unsigned int)ARMY_HAS_HERO);
	for (std::deque<Army*>::iterator it = fighters.begin();
			it != fighters.end(); it++)
		(*it)->set_army_bonus((*it)->get_army_bonus() & compl_mask);
}

// End of file
