//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Library General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include "path.h"
#include <stdlib.h>
#include "army.h"
#include "mapdef.h"
#include "citylist.h"
#include "city.h"
#include "stacklist.h"
#include "defs.h"
#include "stack.h"

//#define debug(x) {std::cerr<<__FILE__<<": "<<__LINE__<<": "<<x<<endl;}
#define debug(x)


//------------------------------------------------------------------------
//Definition of the class PathItem, which is needed by the Path class for
//calculation purposes, but nevertheless very simple and private, so we put it
//here where noone else can see it :)

class PathItem
{
	public:
		//constructors
		PathItem ();	//shouldn't be used, but I don't know yet
		PathItem (int _x, int _y, int _g, int _h, int _parentx, int _parenty);
		PathItem (const PathItem& orig);
		
		//data, public for faster access
		int x, y;
		int g, h;
		int fitness;
		int parentx, parenty;

		//need this operator for creating an automatically sorted list
		bool operator< (const PathItem compare) const;
};


PathItem::PathItem ()
	:x(-1), y(-1), g(0), h(0), fitness(10000), parentx(-1),parenty(-1)
{
}

PathItem::PathItem (int _x, int _y, int _g, int _h, int _parentx, int _parenty)
	:x(_x), y(_y), g(_g), h(_h), parentx(_parentx), parenty(_parenty)
{
	fitness = g+h;
}

PathItem::PathItem (const PathItem& orig)
	:x(orig.x), y(orig.y), g(orig.g), h(orig.h), fitness(orig.fitness),
	 parentx(orig.parentx), parenty(orig.parenty)
{
}

bool PathItem::operator< (const PathItem compare) const
{
	return (fitness < compare.fitness);
}

//--------------------------------------------------------------------------
//now comes the real stuff

typedef std::multiset<PathItem, std::less<PathItem> >::iterator listiterator;
typedef std::map<unsigned int, PathItem, std::less<unsigned int> >::iterator mapiterator;


Path::Path()
{
	set_heuristic(2);
}

Path::~Path()
{
	fl_clear();
}

void Path::set_heuristic (double value)
{
	d_heuristic = value;
}

double Path::get_heuristic ()
{
	return d_heuristic;
}

//calculate the path for the stack. Almost dummy function.
bool Path::calculate (Stack* s, PG_Point _dest)
{
	debug("calculating path...");
	fl_clear();
	calculate_path(s, _dest, s->get_pos(), 0, true);
	debug("...done")
	if (empty()) return false;	//calculation didn't work
	return true;
}

//checks whether path is still valid. This function detects obstacles in the
//path and relies on calculate_path to find a way around them
bool Path::check_path(Stack* s, bool recalculate)
{
	debug("check_path()");
	
	d_has_land = d_has_ship = false;//after all, the path has already been
					//calculated once so we don't need these
	d_unit = s;
	
	d_heuristic *= 2;		//we want straightforward paths
	iterator itstart, itend;
	PG_Point p_start;
	bool blocking = false;
		/* There are 4 possibilities for the boolean variables here:

		- !is_blocked && !blocking
			i.e. tile is not blocked and we aren't currently
			investigating a blocked section
			-> go to next tile

		- !is_blocked && blocking
			i.e. a blocked section has just ended. Mark this point
			and set blocking=false; calculate a workaround

		- is_blocked && !blocking
			i.e. a blocked section has just begun. Mark the
			previous (!!!!!) tile and set blocking = true

		- is_blocked && blocking
			i.e. we are in the middle of a blocked section; nothing
			special with this tile
			->next tile
		*/
			
	for (iterator it = begin(); it != end(); it++)
	{
		bool temp = is_blocked ((*it)->x, (*it)->y);

		if ((!temp) && (!blocking))
		{
			debug((*it)->x<<","<<(*it)->y<<" not blocked\n")
			continue;
		}

		if ((temp) && (blocking))
		{
			debug((*it)->x<<","<<(*it)->y<<" blocked\n")
			it++;
			if (it == end())
				temp = !temp;
			it--;
			continue;
		}

		if ((temp) && (!blocking))
		{
			debug((*it)->x<<","<<(*it)->y<<" blocked\n")
			blocking = true;
			if (it == begin())
			{
				p_start = d_unit->get_pos();
				itstart = it;
				continue;
			}
			itstart = it;
			p_start = **(--itstart);
			continue;
		}

		if ((!temp) && (blocking))
		{
			debug((*it)->x<<","<<(*it)->y<<" not blocked\n")
			if (!recalculate)
			{
				d_heuristic /= 2;
				d_unit = 0;
				return false;
			}

		
			itend = it;
			std::list<PG_Point> templist;
			
			if (!calculate_path (s, *(*itend), p_start,
						&templist, false))
			{
				//no way around obstacle, so clear the useless
				//path and exit
				fl_clear();
				d_heuristic /= 2;
				return false;
			}

			//erase the blocked tiles...
			while (itstart != itend)
				itstart = fl_erase(itstart);

			//...and insert our new ones
			for (std::list<PG_Point>::iterator ittemp = templist.begin();
				ittemp != templist.end(); ittemp++)
			{
				insert(itstart, new PG_Point(*ittemp));
			}

			//done
			continue;
		}
	}

	//path successfully recalculated (or at least not blocked)
	d_heuristic /=2;
	d_unit = 0;
	return true;
}


/* some notes to the A* algorithm used here
   starting from the starting point, we examine all surrounding fields x and give
   them a g(x) and an h(x) value. g(x) is the number of movement points already
   needed and h(x) is calculated by ([distance to destination] * heuristics).
   Heuristics is the average number of MP/field. Then the fitness of each field
   is calculated by f(x) = g(x) + h(x). The lower the fitness, the better the 
   chances we are on the least-cost way (=the lower the estimated MP costs).
   We then put all fields in the open list, take the fittest field from this again
   and redo these steps until we have found the way to the end point. The closed
   list is just here to make sure we don't undo our steps (if there are only
   expensive ways towards the target, you will else get into an endless loop;
   you can ignore it until you have understood how A* works).
   If you want a better explanation, search the internet for "pathfinding + A*" :)
*/

bool Path::calculate_path (Stack* s, PG_Point _dest, PG_Point start,
				std::list<PG_Point>* pathlist, bool write_directly)
{
	debug("calculate_path() from "<<start.x<<","<<start.y<<" to "<<_dest.x<<","<<_dest.y)

	d_dest = _dest;
	d_unit = s;
	d_openlist.clear();
	d_closelist.clear();
	d_army_bonus = d_stack_bonus = !0;
	d_has_ship = d_has_land = false;

	//check if there are ships/non-water units in the stack and
	//calculate the move boni
	for (Stack::iterator it = d_unit->begin(); it != d_unit->end(); it++)
	{
		if ((*it)->get_army_bonus() ==ARMY_IS_LOADED)
			continue;
	
		d_army_bonus &= (*it)->get_army_bonus();
		d_stack_bonus &= (*it)->get_move_bonus();

		if ((*it)->get_army_bonus() == ARMY_SHIP)
		{
			//ship in stack
			d_has_ship = true;
			continue;
		}

		if ((*it)->get_move_bonus() != MAPTILE_WATER)
			//unit which can't cross water
			d_has_land = true;
	}

	if (is_blocked(d_dest.x, d_dest.y))
	{
		d_unit = 0;
		return false;
	}

	//now start the algorithm by inserting the starting position as first
	//piece

	PathItem cur = create_item(start.x, start.y, 0);
	insert_item (cur);

	while (!d_openlist.empty())
	//if openlist becomes empty, there is no way to the destination
	{
		cur = (*d_openlist.begin());
		if (cur.h == 0)		//target reached
			break;
		d_openlist.erase(d_openlist.begin());
		check_surroundings(&cur);
		//put the checked field in closelist
		std::pair<unsigned int, PathItem> pairtemp;
		pairtemp.first = (cur.x << 16) + cur.y;
		pairtemp.second = cur;
		d_closelist.insert(pairtemp);
	}

	d_unit = 0;

	if (d_openlist.empty())
		return false;
	
	if (write_directly)
	//don't pass around the path information via the list
	{
		PG_Point* p = new PG_Point;
		p->x = cur.x;	p->y = cur.y;
		push_front(p);
		while ((cur.x != start.x) || (cur.y != start.y))
		{
			cur = (d_closelist.find((cur.parentx << 16)+cur.parenty))->second;
			p = new PG_Point;
			p->x = cur.x;	p->y = cur.y;
			push_front(p);
		}
		fl_erase(begin());	//remove the first tile (which points
					//at the stack's current location
		return true;
	}

	if ((!write_directly) && (pathlist != 0))
	{
		PG_Point p;
		p.x = cur.x;	p.y = cur.y;
		pathlist->push_front(p);
		while ((cur.x != start.x) || (cur.y != start.y))
		{
			cur = (d_closelist.find((cur.parentx << 16) + cur.parenty))->second;
			p.x = cur.x;	p.y = cur.y;
			pathlist->push_front(p);
		}
		pathlist->erase(pathlist->begin());
		return true;
	}

	//this point shouldn't normally be reached
	return false;
}


void Path::check_surroundings (PathItem* p)
{
	debug("check_surroundings; x="<<p->x<<" y="<<p->y<<endl)
	/*First the following:
	* to improve the path appearance, we divide the environment of the tile
	* in 8 parts like this
	*
	* 		123
	*		8T4
	*		765
	*
	* Now we take the part which looks closest to the target and start going
	* around the tile clockwise.  The idea behind this is in the following
	* situation:

			GGGGG		(S..Start, D..Destination)
			DGGGS
			GGGGG
	
	* If we would always start with the upper left corner, this tile would
	* be preferred (has same fitness as the other left tiles, but it is
	* inserted first => comes first in list). The way resulting from this 
	* would go 2 steps left and up and two steps left and down. This is 
	* stupid, so the algorithm should find out that the destination is
	* left from the start.
	*/

	int dx, dy;

	//if the Destination is on the same x-axis like the tile, direction
	//is 6 or 2
	if (d_dest.x == p->x)
	{
		if (d_dest.y > p->y)
		{
			dx = 0;
			dy = 1;
		}
		if (d_dest.y < p->y)
		{
			dx = 0;
			dy = -1;
		}
	}

	//the same with the y-axis (4 or 8)
	if (d_dest.y == p->y)
	{
		if (d_dest.x > p->x)
		{
			dx = 1;
			dy = 0;
		}
		if (d_dest.x < p->x)
		{
			dx = -1;
			dy = 0;
		}
	}

	//else it is one of the diagonal directions
	if (d_dest.y > p->y)
	{
		if (d_dest.x > p->x)
		{
			dx = dy = 1;
		}
		if (d_dest.x < p->x)
		{
			dx = -1;
			dy = 1;
		}
	}

	if (d_dest.y < p->y)
	{
		if (d_dest.x > p->x)
		{
			dx = 1;
			dy = -1;
		}
		if (d_dest.x < p->x)
		{
			dx = dy = -1;
		}
	}

	//now, let's iterate clockwise
	for (int k = 8; k > 0; k--)
	{
		int xq, yq;
		xq = p->x + dx;
		yq = p->y + dy;
		if ((xq < 100) && (xq >= 0) && (yq < 100)
			&& (yq >= 0) && (!is_blocked(xq, yq)))
		{
			insert_item (create_item (xq, yq, p));	
		}

		//now go to the next cycle (this really works :))
		int temp = dx - dy;
		
		dy += dx;
		if (dy == -2)
			dy = -1;
		if (dy == 2)
			dy = 1;

		dx = temp;
		if (dx == -2)
			dx = -1;
		if (dx == 2)
			dx = 1;
	}
}

bool Path::is_blocked (int x, int y)
{
	//TODO: you can extract quite some amount of time here with a clever
	//search algorithm for stacklist
	
	//return true on every condition which may prevent the stack from
	//entering the tile, which are...
	
	//...land if the stack has ships in it...
	if (d_has_ship && (mapdef->map[x][y]->get_maptile_type() != MAPTILE_WATER)
		&& (mapdef->map[x][y]->get_building() != Maptile::CITY))
		return true;
	
	//...water if some units can't fly or swim...
	if (d_has_land && (mapdef->map[x][y]->get_maptile_type() == MAPTILE_WATER))
		return true;
	
	if (Stack* s = stacklist->get_object_at(x,y))
	{
		//...enemy stacks which stand in the way...
		if ((s->get_player() != d_unit->get_player())
			&& ((x != d_dest.x) || (y != d_dest.y)))
			return true;
	}

	City* c = citylist->get_object_at(x,y);
	//...enemy cities
	if (c && (c->get_player() != d_unit->get_player())
		&& ((x != d_dest.x) || (y != d_dest.y)))
		return true;

	//no obstacles??? well, then...
	return false;
}

PathItem Path::create_item (int x, int y, PathItem* parent)
{
	PathItem item;
	int diffx, diffy, distance;
	
	item.x = x;
	item.y = y;
	item.g = mapdef->map[x][y]->get_moves();
	
	//if move bonus of army covers move bonus of maptile, movement cost is
	//set to 2
	if (mapdef->map[x][y]->get_maptile_type() & d_stack_bonus)
		item.g = 2;
	
	if (parent)
	{
		item.g += parent->g;
		item.parentx = parent->x;
		item.parenty = parent->y;
	}
	else
	{
		item.parentx = item.parenty = -1;
		item.g = 0;	//starting point doesn't need any MP
	}
	
	diffx = abs (d_dest.x - x);
	diffy = abs (d_dest.y - y);

	distance = (diffx > diffy)? diffx : diffy;
		//we can also move diagonal
	
	item.h = (int) (distance * d_heuristic);
	item.fitness = item.g + item.h;

	debug("created maptile x="<<item.x<<" y="<<item.y<<" g="<<item.g<<" h="<<item.h)

	return item;
}

void Path::insert_item (PathItem item)
{
	listiterator it;
	unsigned int mykey = (item.x<<16) + item.y;

	//is item already in closelist, ignore it (BTW: THAT's what the map is
	//for, this is much shorter than iterating)
	if (d_closelist.find(mykey) != d_closelist.end())
		return;

	//if the item is already in openlist, replace it if the new one's cheaper
	for (it = d_openlist.begin(); it != d_openlist.end(); it++)
		if ((item.x == it->x) && (item.y == it->y))
		{
			if (item.fitness < it->fitness)
			{
				d_openlist.erase(it);
				d_openlist.insert(item);
			}
			return;
		}
	
	d_openlist.insert(item);
}


Path::iterator Path::fl_erase(Path::iterator it)
{
	delete (*it);
	return erase (it);
}

void Path::fl_clear()
{
	for (Path::iterator it = begin(); it != end(); it++)
	{
		delete (*it);
	}
	return clear();
}
