//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Library General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include <stdlib.h>
#include <algorithm>
#include "player.h"
#include "playerlist.h"
#include <assert.h>
#include "armysetlist.h"
#include "FightDialog.h"
#include "stacklist.h"
#include "citylist.h"
#include <fstream.h>
#include "path.h"
#ifdef __WIN32__
#include <windows.h>
#endif
#include "mapdef.h"

#define debug(x) {std::cerr<<__FILE__<<": "<<__LINE__<<": "<<x<<endl;}
//#define debug(x)

Player::Player(std::string name, std::string armyset, SDL_Color color)
	:d_color(color), d_name(name), d_armyset(armyset), d_gold(1000),
	d_computer(false), d_dead(false)
{
}

Player::~Player()
{
}

Player::Player(const Player& player)
	:d_color(player.d_color), d_name(player.d_name), d_armyset(player.d_armyset),
	d_gold(player.d_gold), d_computer(player.d_computer), d_dead(player.d_dead),
	d_actions(player.d_actions)
{
}

void Player::addGold(int gold)
{
	d_gold += gold;
	emit(MSG_PLAYER_STATUS_CHANGED);
}

void Player::withdrawGold(int gold)
{
	d_gold -= gold;
	emit(MSG_PLAYER_STATUS_CHANGED);
}

Armyset* Player::get_armyset()
{
	Armyset* armyset = armysetlist->get_armyset(d_armyset);
	assert(armyset != 0);
	return armyset;
}

bool Player::start_turn()
{
	//initialize stuff; until there are directors there are only human
	//players, so just unlock the screen
	emit(MSG_PLAYER_UNLOCK_SCREEN);
	clear_actionlist();

	return true;
}

bool Player::next_turn()
{
	//clean up after AI/network player has finished, up to now there isn't
	//much to do
	emit(MSG_PLAYER_LOCK_SCREEN);

	//start next round, this way creates some stack overhead
	//(Player::next_turn calls Playerlist::next_player which calls
	//Player::start_turn() which may call the director which calls
	//Player::next_turn...), but this is winded down as soon as an actual
	//human player starts his turn (Player::start_turn() exits then)
	playerlist->next_player();

	return true;
}

bool Player::stack_split(Stack* s)
{
	debug("player::split()")

	Army* ungrouped = s->get_first_ungrouped_army();
	if (!ungrouped)		//no armies to split
		return false;
	
	Stack* new_stack = new Stack(this, s->get_pos());

	while (ungrouped)
	{
		new_stack->push_back(ungrouped);
		s->erase(std::find(s->begin(), s->end(), ungrouped));
		
		ungrouped->set_grouped(true);
		ungrouped = s->get_first_ungrouped_army();
	}

	stacklist->push_back(new_stack);

	Action_Split* item = new Action_Split();
	item->fill_data(s, new_stack);
	d_actions.push_back(item);

	emit(MSG_PLAYER_STACK_UPDATE);
	
	return true;
}

bool Player::stack_join(Stack* receiver, Stack* joining)
{
	debug("player::stack_join()")

	if ((receiver == 0) || (joining == 0))
		return false;

	if ((receiver->size() + joining->size()) > 8)
		return false;
	
	for (Stack::iterator it = joining->begin(); it != joining->end(); it++)
		receiver->push_back(*it);

	Action_Join* item = new Action_Join();
	item->fill_data(receiver, joining);
	d_actions.push_back(item);

	joining->clear();	//clear only erases the pointers not the armies
	stacklist->fl_remove(joining);

	emit(MSG_PLAYER_STACK_UPDATE);
	
	return true;
}

bool Player::stack_move(Stack* s, PG_Point dest, bool follow)
{
	//if follow is set to true, follow an already calculated way, else
	//calculate it here
	if (!follow)
		s->get_path()->calculate(s, dest);
	else
		s->get_path()->check_path(s);
	
	if (s->get_path()->empty())
		return false;		//way to destination blocked

	while ((enough_moves(s)) && (s->get_path()->size() > 1))
	{
		stack_move_one_step(s);
		emit(MSG_PLAYER_STACK_UPDATE);
	}

	//If there is another stack where we landed, join it. We can't have two
	//stacks share the same maptile, besides you have e.g. added strength
	//for combat.
	stack_join(s, stacklist->get_ambiguity(s));

	if ((s->get_path()->size() == 1) && enough_moves(s))
	//now look for fight targets, joins etc.
	{
		PG_Point pos = **(s->get_path()->begin());
		City* city = citylist->get_object_at(pos);
		Stack* target = stacklist->get_object_at(pos);

		//first fight_city to avoid ambiguity with fight_army
		if (city && (city->get_player() != this))
		{
			Fight::RESULT result = stack_fight_city
						(std::vector<Stack*>(1, s), city);

			if (result == Fight::ATTACKER_WON)
			{
				emit(MSG_PLAYER_CITY_UPDATE);
				city->set_player(this);
				stack_move_one_step(s);
			}
			if (result == Fight::DEFENDER_WON)
			{
				emit(MSG_PLAYER_STACK_DIED);
			}
			return true;
		}
		else if (target)
		{
		}

	}

	return true;
}

Fight::RESULT Player::stack_fight(Stack* sattacker, Stack* sdefender)
{
	debug("player::fight()")

	if ((sattacker == 0) || (sdefender == 0))
		return Fight::ATTACKER_WON;

	//first of all, copy the stacks (we need the original ones to compare
	//the wounded and killed units
	Stack* attacker = new Stack(*sattacker);
	Stack* defender = new Stack(*sdefender);
	Stack::iterator it;

	//initialize and do the Fight
	FightDialog fdlg(sattacker, sdefender, 0);
	fdlg.Show();
	fdlg.battle();

	Fight::RESULT result = fdlg.get_result();

	//add result to actionlist
	Action_Fight* item = new Action_Fight;
	item->fill_data(sattacker, sdefender, attacker, defender);
	d_actions.push_back(item);

	//now clean up the stacklist if necessary
	delete defender;
	delete attacker;

	for (it = sattacker->begin(); it != sattacker->end();)
	{
		if ((*it)->get_hp() <= 0)
		{
			it = sattacker->fl_erase(it);
			continue;
		}
		it++;
	}
	if (sattacker->empty())
		//stack destroyed, so remove it from the stacklist
		stacklist->fl_remove(sattacker);

	//now the same procedure for the defender stack
	for (it = sdefender->begin(); it != sdefender->end();)
	{
		if ((*it)->get_hp() <= 0)
		{
			it = sdefender->erase(it);
			continue;
		}
		it++;
	}
	if (sdefender->empty())
		stacklist->fl_remove(sdefender);

	emit(MSG_PLAYER_STACK_UPDATE);

	return result;
}

Fight::RESULT Player::stack_fight_city(std::vector<Stack*> attackers, 
						City* target)
{
	debug("player::stack_fight_city()")

	std::vector<Stack*> defenders = stacklist->defenders_in_city(target);
	if (defenders.empty())
		return Fight::ATTACKER_WON;

	//first of all, copy the Stacks, for comparison later on
	std::vector<Stack*> orig_attackers, orig_defenders;
	std::vector<Stack*>::iterator it;

	for (it = attackers.begin(); it != attackers.end(); it++)
	{
		Stack* s = new Stack(**it);
		orig_attackers.push_back(s);
	}
	for (it = defenders.begin(); it != defenders.end(); it++)
	{
		Stack* s = new Stack(**it);
		orig_defenders.push_back(s);
	}

	//now launch the fight
	FightDialog fdlg(attackers, defenders, 0, 0);
	fdlg.Show();
	fdlg.battle();

	Fight::RESULT result = fdlg.get_result();

	//add entry to the actionlist
	Action_Fight_City* item = new Action_Fight_City();
	item->fill_data(orig_attackers, orig_defenders, attackers, defenders);
	d_actions.push_back(item);

	//remove dead armies
	//small shortcut (if you attack a city you are usually strong enough
	//to occupy it)
	if (result == Fight::ATTACKER_WON)
	{
		for (it = defenders.begin(); it != defenders.end(); it++)
			stacklist->fl_remove(*it);
		defenders.clear();
	}
	else
	{
		for (it = defenders.begin(); it != defenders.end(); it++)
		{
			for (Stack::iterator sit = (*it)->begin();
				sit != (*it)->end();)
			{
				if ((*sit)->get_hp() <= 0)
				{
					sit = (*it)->fl_erase(sit);
					continue;
				}
				sit++;
			}
			if ((*it)->empty())
			{
				stacklist->fl_remove(*it);
				(*it) = 0;
			}
		}
	}

	//now clear up attackers like this
	if (result == Fight::DEFENDER_WON)
	{
		for (it = attackers.begin(); it != attackers.end(); it++)
			stacklist->fl_remove(*it);
		attackers.clear();

		emit(MSG_PLAYER_STACK_DIED);
	}
	else
	{
		for (it = attackers.begin(); it != attackers.end(); it++)
		{
			for (Stack::iterator sit = (*it)->begin();
				sit != (*it)->end();)
			{
				if ((*sit)->get_hp() <= 0)
				{
					sit = (*it)->fl_erase(sit);
					continue;
				}
				sit++;
			}
			if ((*it)->empty())
			{
				stacklist->fl_remove(*it);
				(*it) = 0;
			}
		}
		emit(MSG_PLAYER_STACK_UPDATE);
	}

	return result;
}

bool Player::stack_search_ruin(Stack* s, Ruin* r)
{
	debug("Player::stack_search_ruin")
	
	//throw out impossible actions
	if ((s->get_pos().x != r->get_pos().x) ||
		(s->get_pos().y != r->get_pos().y))
	{
		std::cerr <<"error, searching stack and ruin to be searched";
		std::cerr <<"not on same position\n";
		exit(-1);
	}

	if (r->isSearched())
		return false;

	Stack* keeper = r->get_occupant();
	Stack* o_keeper = new Stack(*keeper);
	Stack* o_explorer = new Stack(*s);
	
	//now the stack has to fight the keeper
	if (r->get_occupant() != 0)
	{
		FightDialog fdlg(s, keeper, 0, false);
		fdlg.Show();
		fdlg.battle();

		Fight::RESULT result = fdlg.get_result();

		//searcher didn't kill occupant stack, so ruin is not searched
		if (result != Fight::ATTACKER_WON)
		{
			Action_Search* item = new Action_Search();
			item->fill_data(r, o_keeper, o_explorer, s, false);
			d_actions.push_back(item);
		}
		else
		{
			Action_Search* item = new Action_Search();
			item->fill_data(r, o_keeper, o_explorer, s, true);
			d_actions.push_back(item);
		}

		//now delete the help stacks and remove stiffs
		delete o_keeper;
		delete o_explorer;

		for (Stack::iterator it = s->begin(); it != s->end();)
		{
			if ((*it)->get_hp() <= 0)
			{
				it = s->fl_erase(it);
				continue;
			}
			it++;
		}
		if (s->empty())
		{
			stacklist->fl_remove(s);
			s = 0;
		}

		//if attacker has not succeded in killing the defender, remove
		//dead defenders and quit here
		if (result != Fight::ATTACKER_WON)
		{
			for (Stack::iterator it = keeper->begin(); it != keeper->end();)
			{
				if ((*it)->get_hp() <= 0)
				{
					it =keeper->fl_erase(it);
					continue;
				}
				it++;
			}

			if (!s)
				emit(MSG_PLAYER_STACK_DIED);
			else
				emit(MSG_PLAYER_STACK_UPDATE);
			

			return false;
		}

		r->set_occupant(0);
		delete keeper;
	}
	else		//r->get_occupant() == 0
	{
		Action_Search* item = new Action_Search();
		item->fill_data(r, 0, 0, 0, true);
		d_actions.push_back(item);
	}
	
	//The fight has been done or left out, now comes the reward. Up to now,
	//we only give some gold (automatically done with give_reward())
	r->setSearched(true);
	give_reward(1, s);

	emit(MSG_PLAYER_STACK_UPDATE);

	return true;
}

bool Player::stack_visit_temple(Stack* s, Temple* t)
{
	debug("Player::visit_temple")

	//abort in case of impossible action
	if ((t->get_pos().x != s->get_pos().x)
		|| (t->get_pos().y != s->get_pos().y))
	{
		std::cerr <<"Stack tried to visit temple at wrong location\n";
		exit(-1);
	}

	//later on, you can also receive quests in temples, but for now it
	//must suffice to have your stack blessed (+1 strength)
	s->bless();

	Action_Temple* item = new Action_Temple();
	item->fill_data(t, s);
	d_actions.push_back(item);

	emit(MSG_PLAYER_STACK_UPDATE);

	return true;
}

bool Player::city_occupy(City* c)
{
	c->set_player(this);

	Action_Occupy* item = new Action_Occupy();
	item->fill_data(c);
	d_actions.push_back(item);

	//to signal that the cities have changed a bit
	emit(MSG_PLAYER_CITY_UPDATE);

	return true;
}

bool Player::city_raze(City* c)
{
	//TBD: don't erase the city, but burn it down (we're among other things
	//missing a pic for it)
	Action_Raze* item = new Action_Raze();
	item->fill_data(c);
	d_actions.push_back(item);

	Citylist::iterator it = citylist->begin();
	while ((&(*it) != c) && (it != citylist->end()))
		it++;
	citylist->erase(it);

	emit(MSG_PLAYER_CITY_UPDATE);

	return true;
}

bool Player::city_upgrade_defense(City* c)
{
	debug("Player::city_upgrade_defense")

	int needed_gold = c->get_defense_level() * 1000;

	if ((needed_gold >= 3000) || (d_gold < needed_gold))
		return false;	//maximum level is 3 or not enough gold
	
	c->set_defense_level(c->get_defense_level() + 1);
	withdrawGold(needed_gold);

	Action_Upgrade_Defense* item = new Action_Upgrade_Defense();
	item->fill_data(c);
	d_actions.push_back(item);

	return true;
}

bool Player::city_buy_production(City* c, int armytype)
{
	int slot = -1;

	if ((armytype < 0) || (armytype > 11))
		return false;	//wrong production type (from 0 to 11)

	for (int i = 0; i < 4; i++)
	{
		if ((slot == -1) && (c->get_armytype(i) == -1))
			slot = i;
		if (c->get_armytype(i) == armytype)
			return false;	//city already has this production
	}
	
	if (slot == -1)
		return false;	//city already has all possible production slots
	
	int needed_gold = get_armyset()->army[armytype].get_production_cost();

	if (needed_gold > d_gold)
		return false;	//not enough gold
	
	
	withdrawGold(needed_gold);
	c->set_armytype(slot, armytype);

	Action_Buy_Production* item = new Action_Buy_Production();
	item->fill_data(c, armytype);
	d_actions.push_back(item);

	return true;
}

bool Player::city_change_production(City* c, int production)
{
	if (production > 4)
		return false;	//cities currently have only 4 slots
	
	if (c->get_armytype(production) == -1)
		return false;	//the demanded production slot doesn't exist
	
	c->set_production(production);

	Action_Change_Production* item = new Action_Change_Production();
	item->fill_data(c, production);
	d_actions.push_back(item);

	return true;
}

void Player::dump_actionlist()
{
	for (std::list<Action*>::iterator it = d_actions.begin();
		it != d_actions.end(); it++)
	{
		std::cerr <<(*it)->dump() <<endl;
	}	
}

void Player::clear_actionlist()
{
	for (std::list<Action*>::iterator it = d_actions.begin();
		it != d_actions.end(); it++)
		delete (*it);
	d_actions.clear();
}

bool Player::give_reward(int level, Stack* s, fl_Treasure* reward)
{
	debug("Player::give_reward")

	//by now we don't have the infrastructure to give something different
	//from gold, so this function isn't great

	int gold_amount = rand() % 1000;
	addGold(gold_amount);

	fl_Treasure treasure;
	treasure.type = TREASURE_GOLD;
	treasure.subtype = 0;
	treasure.amount = gold_amount;

	Action_Reward* item = new Action_Reward();
	item->fill_data(0, treasure);
	d_actions.push_back(item);

	return true;
}

std::string Player::getAICommand()
{
	std::string command;  
#ifdef __WIN32__
    WinExec("c:/Python22/pythonw.exe ai/easy.py", SW_HIDE);
#else
    system("ai/easy.py");
#endif

    std::ifstream aicommand("aicommand"); // open citydatabase
    aicommand >> command;

    debug("AI-Command = " << command);    
    return command;
}

void Player::computerTurn()
{
    /* This is how scripting should work (this code actualy really works ;))
    std::string command;
    while (command != "ENDTURN")
    {
        command = getAICommand();

        // do all different computer actions here
    
        if (command == "MOVE")
        {
            Stack* s = stacklist->set_next();
            PG_Point p;
            p.x = p.y = 10;
            s->get_path()->calculate(s, p);
            stacklist->move(s);
        }
    }
    */
    
    // easy AI player, but very fast!
	for (Stacklist::iterator it = stacklist->begin(); it != stacklist->end(); ++it)
	{
	    stacklist->set_activestack(*it);
	    Stack* s = *it;
		if (s->get_player() == this && s->size() >= 8)
        {
             PG_Point p = citylist->get_nearest_enemy_city(s->get_pos())->get_pos();
             s->get_path()->calculate(s, p);
             stacklist->move(s);
        }
        // a stack has died -> further iterating leads to a crash so we 
        // start again (recursive programming, therefore the break
        // afterwards is _VERY_ important)
        if (!stacklist->get_activestack())
        {
            computerTurn();
            break;
        }
	}
}

bool Player::stack_move_one_step(Stack* s)
{
	if (!enough_moves(s))
		return false;
	
	PG_Point dest = *(s->get_path()->front());
	
	int needed_moves = mapdef->map[dest.x][dest.y]->get_moves();
	unsigned int maptype = mapdef->map[dest.x][dest.y]->get_maptile_type();

	for (Stack::iterator it = s->begin(); it != s->end(); it++)
	//calculate possible move boni for each army
	{
		if ((*it)->get_move_bonus() == maptype)
		//if army has move bonus, it takes only 2 move points...
		{
			(*it)->decrement_moves(2);
			continue;
		}
		//else the whole
		(*it)->decrement_moves(needed_moves);
	}

	s->move_one_step();	//this is only for updating positions etc.

	Action_Move_Step* item = new Action_Move_Step();
	item->fill_data(s, dest);
	d_actions.push_back(item);
	
	return true;
}

bool Player::enough_moves(Stack* s)
{
	PG_Point dest = *(s->get_path()->front());
	unsigned int maptype = mapdef->map[dest.x][dest.y]->get_maptile_type();
	int needed_moves = mapdef->map[dest.x][dest.y]->get_moves();
	int cur_moves;

	for (Stack::iterator it = s->begin(); it != s->end(); it++)
	{
		cur_moves = needed_moves;
		if ((*it)->get_move_bonus() == maptype)
			cur_moves = 2;

		if ((*it)->get_moves() < cur_moves)
			return false;
	}

	return true;
}

// End of file
