//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Library General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include "bigmap.h"
#include <SDL_image.h>
#include <assert.h>
#include "cityinfo.h"
#include "army.h"
#include "armyset.h"
#include "path.h"
#include "mapdef.h"
#include "stacklist.h"
#include "citylist.h"
#include "ruin.h"
#include "temple.h"
#include "playerlist.h"
#include "defs.h"

//#define debug(x) {std::cerr<<__FILE__<<": "<<__LINE__<<": "<<x<<endl;}
#define debug(x)

BigMap::BigMap(PG_Widget* parent, PG_Rect rect)
	:PG_Widget(parent, rect, true), d_viewrect(0)
{
	d_enable = true;

	// load all pictures
	d_ruinpic = IMG_Load("../pic/map/default/ruin.png");
	assert(d_ruinpic);
	d_templepic = IMG_Load("../pic/map/default/temple.png");
	d_swordpic = IMG_Load("../pic/various/sword.png");
	d_selector[0] = IMG_Load("../pic/various/select0000.png");
	d_selector[1] = IMG_Load("../pic/various/select0001.png");
	d_polepic = IMG_Load("../pic/various/flag_pole.png");

	char str[80];
	for (int i = 0; i < 9; i++)
	{
		sprintf(str, "../pic/map/default/small_castle_%i.png", i);
		d_citypic[i] = IMG_Load(str);
		for (int j = 0; j < 8; j++)
		{
			sprintf(str, "../pic/various/flag_%i_%i.png", i, j+1);
			d_flagpic[i][j] = IMG_Load(str);
		}
	}

	d_citypic[8] = IMG_Load("../pic/map/default/small_castle_neutral.png");

	// creating a surface with the statical parts of the map as maptiles,
	// cities, ruins & temples looks like this is going to be a BIG memory hog
	// (after inspection I found out that this takes around 55MB memory !!!)
	
	SDL_PixelFormat* fmt = GetScreenSurface()->format;
	d_static_map = SDL_CreateRGBSurface(SDL_SWSURFACE, 5000, 5000,
			fmt->BitsPerPixel, fmt->Rmask, fmt->Gmask,
			fmt->Bmask, fmt->Amask);

	// Draw map
	for(int i = 0; i < 100; i++)
	{
		for(int j = 0; j < 100; j++)
		{
			// temphack, needed for the current maptiles
			SDL_Surface* s = (*mapset)[0]->get_pixmap(5, 0);
			assert(s);
			SDL_BlitSurface(s, 0, d_static_map,
					PG_Rect(i*50, j*50, 50, 50).SDLRect());
			// ---

			s = mapdef->map[i][j]->get_picture();
			assert(s);
			SDL_BlitSurface(s, 0, d_static_map,
					PG_Rect(i*50, j*50, 50, 50).SDLRect());
		}
	}

	// Draw all ruins
	for (Ruinlist::iterator it = ruinlist->begin(); it != ruinlist->end(); ++it)
	{
		PG_Point p = (*it).get_pos();
		SDL_BlitSurface(d_ruinpic, 0, d_static_map, PG_Rect(p.x*50, p.y*50, 50, 50).SDLRect());
	}

	// Draw temples
	for (Templelist::iterator it = templelist->begin(); it != templelist->end(); ++it)
	{
		PG_Point p = (*it).get_pos();
		SDL_BlitSurface(d_templepic, 0, d_static_map, PG_Rect(p.x*50, p.y*50, 50, 50).SDLRect()); 
	}
}

BigMap::~BigMap()
{
    SDL_FreeSurface(d_static_map);
    SDL_FreeSurface(d_ruinpic);
    SDL_FreeSurface(d_templepic);
    SDL_FreeSurface(d_swordpic);
    SDL_FreeSurface(d_selector[0]);
    SDL_FreeSurface(d_selector[1]);
    SDL_FreeSurface(d_polepic);

	for (int i = 0; i < 9; i++)
	{
	    SDL_FreeSurface(d_citypic[i]);
		for (int j = 0; j < 8; j++)
		{
            SDL_FreeSurface(d_flagpic[i][j]);
		}
	}
}

void BigMap::eventDraw(SDL_Surface* surface, const PG_Rect& rect)
{
	debug("eventDraw()");

	SDL_BlitSurface(d_static_map, PG_Rect(d_viewrect->my_xpos*50, d_viewrect->my_ypos*50,
				d_viewrect->my_width*50, d_viewrect->my_height*50).SDLRect(), surface, 0);

	DrawBorder(rect, 1);

	// Draw all cities
	for (Citylist::iterator it = citylist->begin(); it != citylist->end(); ++it)
	{	
		PG_Point p = (*it).get_pos();
		if (d_viewrect->IsInside(p))
		{
			p = get_relative_pos(p);

			// set picture for normal or neutral player
			int nr = playerlist->get_nr_of_player((*it).get_player());
			if (nr == -1) nr = 8;
			SDL_BlitSurface(d_citypic[nr], 0, surface, PG_Rect(p.x, p.y, 100, 100).SDLRect());
		}
	}

	// Draw stacks
	for (Stacklist::iterator it = stacklist->begin(); it != stacklist->end(); it++)
	{
		PG_Point p = (*it)->get_pos();
		
		// check if the object lies in the viewed part of the map
		// otherwise we shouldn't draw it
		if (d_viewrect->IsInside(p))
		{
			p = get_relative_pos(p);

			// draw stack
			SDL_BlitSurface((*it)->get_strongest_army()->get_pixmap(), 0,
					surface, PG_Rect(p.x + 6, p.y + 6, 40, 40).SDLRect());

			// draw flagpole
			SDL_BlitSurface(d_polepic, 0, surface, PG_Rect(p.x, p.y, 50, 50).SDLRect());

			// draw flag
			int nr = playerlist->get_nr_of_player((*it)->get_player());
			if (nr == -1) nr = 8;	// for neutral player

			if (d_flagpic[nr][(*it)->size()-1]) 
			{
				SDL_BlitSurface(d_flagpic[nr][(*it)->size()-1], 0, surface,
						PG_Rect(p.x, p.y, 50, 50).SDLRect());
			}
		}
	}

	Stack* stack = stacklist->get_activestack();
	if (!stack) return;

	// Draw Path
	if (stack->get_path()->size())
	{
		SDL_Color c;
		c.r = c.g = c.b = 0;

		for (std::list<PG_Point*>::iterator it = stack->get_path()->begin();
			it != --(stack->get_path()->end()); ++it)
		{
			PG_Point pos = get_relative_pos(**it);
			DrawLine(pos.x + 15, pos.y + 15, pos.x + 30, pos.y + 30, c);
			DrawLine(pos.x + 15, pos.y + 30, pos.x + 30, pos.y + 15, c);
		}

		PG_Point pos = get_relative_pos(*stack->get_path()->back());
		SDL_BlitSurface(d_swordpic, 0, surface,
				PG_Rect(pos.x + 5, pos.y + 5, 40, 40).SDLRect());
	}

	// Draw Selector
	PG_Point p = stack->get_pos();
	if (d_viewrect->IsInside(p))
	{
		p = get_relative_pos(p);
		SDL_BlitSurface(d_selector[0], 0, surface, PG_Rect(p.x, p.y, 50, 50).SDLRect());
	}
}

bool BigMap::eventMouseButtonDown(const SDL_MouseButtonEvent* event)
{
	if (!d_enable)
		//pretend to handle event, but don't do it (right now we're
		//waiting for the AI or network guy)
		return true;

	int pos_X = (event->x - my_xpos) / 50 + d_viewrect->my_xpos;
	int pos_Y = (event->y - my_ypos) / 50 + d_viewrect->my_ypos;

	if (event->button == SDL_BUTTON_LEFT)
	{
		Stack* stack = stacklist->get_activestack();

		if (stack)
		{
			// clicked on an already active stack
			if (stack->get_pos().x == pos_X && stack->get_pos().y == pos_Y) return true;

			PG_Point p;
			p.x = pos_X; p.y = pos_Y;

			playerlist->get_activeplayer()->stack_split(stack); // split if ungrouped
			stack->get_path()->calculate(stack, p);
			Redraw();
			stacklist->move(stack);
		}

		// Stack hasn't been active yet
		else
		{
			stack = stacklist->get_object_at(pos_X, pos_Y);
			if ((stack) && (stack->is_friend(playerlist->get_activeplayer()))) 
			{
				stacklist->set_activestack(stack);
				stack_selected();
			}

		}
	}

	// right mousebutton to click at city
	if (event->button == SDL_BUTTON_RIGHT)
	{
		City* city = citylist->get_object_at(pos_X, pos_Y);

		if ((city) && (city->is_friend(playerlist->get_activeplayer())))
		{
			CityInfo cityinfo(city);
			cityinfo.Show();
			cityinfo.RunModal();
			cityinfo.Hide();
		}
	}

	return true;
}

void BigMap::center_view(const PG_Point& p)
{
	debug("center_view()");

	d_viewrect->my_xpos = p.x - (int)(d_viewrect->my_width / 2);
	d_viewrect->my_ypos = p.y - (int)(d_viewrect->my_height / 2);

	if (d_viewrect->my_xpos < 0) d_viewrect->my_xpos = 0;
	if (d_viewrect->my_ypos < 0) d_viewrect->my_ypos = 0;
	if (d_viewrect->my_xpos > (100 - d_viewrect->my_width)) d_viewrect->my_xpos = 100 - d_viewrect->my_width;
	if (d_viewrect->my_ypos > (100 - d_viewrect->my_height)) d_viewrect->my_ypos = 100 - d_viewrect->my_height;

	emit(MSG_BIGMAP_MOVED_VIEWRECT);
}

void BigMap::stack_moved()
{
	debug("stack_moved()");

	Stack* stack = stacklist->get_activestack();
	assert(stack != 0);

	center_view(stack->get_pos());  // center the view around the new stack position
	Redraw();
}

void BigMap::stack_selected()
{
	debug("stack_selected()");

	Stack* stack = stacklist->get_activestack();
	if (!stack->get_path()->check_path(stack))
	{
		//error handling is required here, up to now we only barf on cerr
		cerr << "original path of stack was blocked\n";
	}

	center_view(stack->get_pos());
	Redraw();

	emit(MSG_BIGMAP_STACK_SELECTED);
}

void BigMap::stack_deselected()
{
	debug("stack_deselected()");
	stacklist->set_activestack(0);
	Redraw();
}

// calculate the relative pos from the absolute pos by subtracting
// the viewrect-pos and multiplying with MAPTILE_SIZE (50)

const PG_Point BigMap::get_relative_pos(const PG_Point& absolute_pos)
{
	PG_Point relative_pos;
	relative_pos.x = (absolute_pos.x - d_viewrect->my_xpos) * 50;
	relative_pos.y = (absolute_pos.y - d_viewrect->my_ypos) * 50;
	return relative_pos;
}

// End of file
